@@ -1,4 +1,5 @@
 import argparse
+import logging
 import re
 import time
 from collections import defaultdict
@@ -8,17 +9,19 @@ from typing import Dict, List, Optional, Set, Tuple
 import numpy as np
 import omero.clients  # noqa
 from ome_zarr.conversions import int_to_rgba_255
-from ome_zarr.data import write_multiscale
 from ome_zarr.io import parse_url
 from ome_zarr.reader import Multiscales, Node
 from ome_zarr.scale import Scaler
 from ome_zarr.types import JSONDict
+from ome_zarr.writer import write_multiscale_labels
 from omero.model import MaskI, PolygonI
 from omero.rtypes import unwrap
 from skimage.draw import polygon as sk_polygon
-from zarr.convenience import open as zarr_open
+from zarr.hierarchy import open_group
 
-from .util import print_status
+from .util import marshal_axes, marshal_transformations, open_store, print_status
+
+LOGGER = logging.getLogger("omero_zarr.masks")
 
 # Mapping of dimension names to axes in the Zarr
 DIMENSION_ORDER: Dict[str, int] = {
@@ -59,7 +62,13 @@ def plate_shapes_to_zarr(
 
     dtype = MASK_DTYPE_SIZE[int(args.label_bits)]
     saver = MaskSaver(
-        plate, None, dtype, args.label_path, args.style, args.source_image
+        plate,
+        None,
+        dtype,
+        args.label_path,
+        args.style,
+        args.source_image,
+        args.overlaps,
     )
 
     count = 0
@@ -151,7 +160,13 @@ def image_shapes_to_zarr(
 
     if masks:
         saver = MaskSaver(
-            None, image, dtype, args.label_path, args.style, args.source_image
+            None,
+            image,
+            dtype,
+            args.label_path,
+            args.style,
+            args.source_image,
+            args.overlaps,
         )
 
         if args.style == "split":
@@ -175,6 +190,8 @@ class MaskSaver:
     masks to zarr groups/arrays.
     """
 
+    OVERLAPS = ("error", "dtype_max")
+
     def __init__(
         self,
         plate: Optional[omero.gateway.PlateWrapper],
@@ -183,6 +200,7 @@ class MaskSaver:
         path: str = "labels",
         style: str = "labeled",
         source: str = "..",
+        overlaps: str = "error",
     ) -> None:
         self.dtype = dtype
         self.path = path
@@ -190,6 +208,7 @@ class MaskSaver:
         self.source_image = source
         self.plate = plate
         self.plate_path = Optional[str]
+        self.overlaps = overlaps
         if image:
             self.image = image
             self.size_t = image.getSizeT()
@@ -215,6 +234,7 @@ class MaskSaver:
         :param plate_path: The zarr path to the image
         :return: None
         """
+        self.image = image
         self.size_t = image.getSizeT()
         self.size_c = image.getSizeC()
         self.size_z = image.getSizeZ()
@@ -267,9 +287,6 @@ class MaskSaver:
         if self.plate:
             assert self.plate_path, "Need image path within the plate"
             source_image = f"{source_image}/{self.plate_path}"
-            current_path = f"{self.plate_path}/{self.path}"
-        else:
-            current_path = self.path
 
         print(f"source_image {source_image}")
         src = parse_url(source_image)
@@ -278,12 +295,13 @@ class MaskSaver:
         assert input_pyramid.load(Multiscales), "No multiscales metadata found"
         input_pyramid_levels = len(input_pyramid.data)
 
-        root = zarr_open(filename)
+        store = open_store(filename)
+        root = open_group(store)
 
-        if current_path in root.group_keys():
-            out_labels = getattr(root, current_path)
+        if self.plate:
+            label_group = root.require_group(self.plate_path)
         else:
-            out_labels = root.require_group(current_path)
+            label_group = root
 
         _mask_shape: List[int] = list(self.image_shape)
         for d in ignored_dimensions:
@@ -297,19 +315,29 @@ class MaskSaver:
 
         # Create and store binary data
         labels, fill_colors, properties = self.masks_to_labels(
-            masks, mask_shape, ignored_dimensions, check_overlaps=True,
+            masks,
+            mask_shape,
+            ignored_dimensions,
         )
+
+        axes = marshal_axes(self.image)
+
+        # For v0.3+ ngff we want to reduce the number of dimensions to
+        # match the dims of the Image.
+        dims_to_squeeze = []
+        for dim, size in enumerate(self.image_shape):
+            if size == 1:
+                dims_to_squeeze.append(dim)
+        labels = np.squeeze(labels, axis=tuple(dims_to_squeeze))
+
         scaler = Scaler(max_layer=input_pyramid_levels)
         label_pyramid = scaler.nearest(labels)
-        pyramid_grp = out_labels.require_group(name)
-
-        write_multiscale(label_pyramid, pyramid_grp)  # TODO: dtype, chunks, overwite
+        transformations = marshal_transformations(self.image, levels=len(label_pyramid))
 
         # Specify and store metadata
         image_label_colors: List[JSONDict] = []
         label_properties: List[JSONDict] = []
         image_label = {
-            "version": "0.2",
             "colors": image_label_colors,
             "properties": label_properties,
             "source": {"image": source_image_link},
@@ -323,19 +351,15 @@ class MaskSaver:
                 image_label_colors.append(
                     {"label-value": label_value, "rgba": int_to_rgba_255(rgba_int)}
                 )
-        # TODO: move to write method
-        pyramid_grp.attrs["image-label"] = image_label
 
-        # Register with labels metadata
-        print(f"Created {filename}/{current_path}/{name}")
-        attrs = out_labels.attrs.asdict()
-        # TODO: could temporarily support "masks" here as well
-        if "labels" in attrs:
-            if name not in attrs["labels"]:
-                attrs["labels"].append(name)
-        else:
-            attrs["labels"] = [name]
-        out_labels.attrs.update(attrs)
+        write_multiscale_labels(
+            label_pyramid,
+            label_group,
+            name,
+            axes=axes,
+            coordinate_transformations=transformations,
+            label_metadata=image_label,
+        )
 
     def shape_to_binim_yx(
         self, shape: omero.model.Shape
@@ -424,7 +448,7 @@ class MaskSaver:
         masks: List[omero.model.Mask],
         mask_shape: Tuple[int, ...],
         ignored_dimensions: Set[str] = None,
-        check_overlaps: bool = True,
+        check_overlaps: bool = None,
     ) -> Tuple[np.ndarray, Dict[int, str], Dict[int, Dict]]:
         """
         :param masks [MaskI]: Iterable container of OMERO masks
@@ -451,7 +475,30 @@ class MaskSaver:
         size_z: int = mask_shape[2]
         ignored_dimensions = ignored_dimensions or set()
 
-        labels = np.zeros(mask_shape, np.int64)
+        roi_ids = [shape.roi.id.val for mask in masks for shape in mask]
+        sorted_roi_ids = list(set(roi_ids))
+        sorted_roi_ids.sort()
+
+        if check_overlaps is None:
+            # If overlaps isn't 'dtype_max', an exception is thrown
+            # if any overlaps exist
+            check_overlaps = self.overlaps != "dtype_max"
+
+        # label values are 1...n
+        max_value = len(sorted_roi_ids)
+        # find most suitable dtype...
+        labels_dtype = np.int64
+        sorted_dtypes = [kv for kv in MASK_DTYPE_SIZE.items()]
+        sorted_dtypes.sort(key=lambda x: x[0])
+        # ignore first dtype (bool)
+        for int_dtype in sorted_dtypes[1:]:
+            dtype = int_dtype[1]
+            # choose first dtype that handles max_value
+            if np.iinfo(dtype).max >= max_value:
+                labels_dtype = dtype
+                break
+        LOGGER.debug("Exporting labels to dtype %s" % labels_dtype)
+        labels = np.zeros(mask_shape, labels_dtype)
 
         for d in "TCZYX":
             if d in ignored_dimensions:
@@ -470,7 +517,7 @@ class MaskSaver:
                 # Using ROI ID allows stitching label from multiple images
                 # into a Plate and not creating duplicates from different iamges.
                 # All shapes will be the same value (color) for each ROI
-                shape_value = shape.roi.id.val
+                shape_value = sorted_roi_ids.index(shape.roi.id.val) + 1
                 properties[shape_value] = {
                     "omero:shapeId": shape.id.val,
                     "omero:roiId": shape.roi.id.val,
@@ -485,21 +532,27 @@ class MaskSaver:
                         for i_z in self._get_indices(
                             ignored_dimensions, "Z", z, size_z
                         ):
-                            if check_overlaps and np.any(
-                                np.logical_and(
-                                    labels[
-                                        i_t, i_c, i_z, y : (y + h), x : (x + w)
-                                    ].astype(np.bool),
-                                    binim_yx,
-                                )
-                            ):
-                                raise Exception(
-                                    f"Mask {shape_value} overlaps with existing labels"
-                                )
+                            overlap = np.logical_and(
+                                labels[i_t, i_c, i_z, y : (y + h), x : (x + w)].astype(
+                                    np.bool
+                                ),
+                                binim_yx,
+                            )
                             # ADD to the array, so zeros in our binarray don't
                             # wipe out previous shapes
                             labels[i_t, i_c, i_z, y : (y + h), x : (x + w)] += (
                                 binim_yx * shape_value
                             )
 
+                            if np.any(overlap):
+                                if check_overlaps:
+                                    raise Exception(
+                                        f"Shape {shape.roi.id.val} overlaps "
+                                        "with existing labels"
+                                    )
+                                else:
+                                    # set overlapping region to max(dtype)
+                                    labels[i_t, i_c, i_z, y : (y + h), x : (x + w)][
+                                        overlap
+                                    ] = np.iinfo(labels_dtype).max
         return labels, fillColors, properties
